#include <petscksp.h>

static char help[] = 
"Reads in binary files for matrix A and RHS vector b. Solves the      \n\
following linear system  of equations:                                \n\
                                                                      \n\
  Ax = b                                                              \n\
                                                                      \n\
Input parameters are:                                                 \n\
                                                                      \n\
  -A <filename> : input matrix in native PETSc format (REQUIRED)      \n\
  -b <filename> : input vector in native PETSc format (REQUIRED)      \n\
  -x_file <filename>: text file to save solution vector to (OPTIONAL) \n\
  -x_view : flag for viewing the solution vector (OPTIONAL)           \n\
                                                                      \n";

int main(int argc, char **args)
{
  Mat             A;
  Vec             b,x;
  KSP             ksp;
  PetscBool       flg_A,flg_b;
  PetscBool       flg_xview=PETSC_FALSE,flg_xfile=PETSC_FALSE;
  PetscViewer     fd;
  PetscErrorCode  ierr;
  char            Ain[PETSC_MAX_PATH_LEN],rhs[PETSC_MAX_PATH_LEN];
  char            xfile[PETSC_MAX_PATH_LEN] = "x_file";

  /* Initialize PETSc program */
  ierr = PetscInitialize(&argc,&args,(char*)0,help);CHKERRQ(ierr);

  /* Required input */
  ierr = PetscOptionsGetString(NULL,NULL,"-A",Ain,PETSC_MAX_PATH_LEN,&flg_A);
  ierr = PetscOptionsGetString(NULL,NULL,"-b",rhs,PETSC_MAX_PATH_LEN,&flg_b);
  if (!flg_A || !flg_b) SETERRQ(PETSC_COMM_WORLD,1,"Must pass in arguments -A <filename> and -b <filename>");
  
  /* Optional input */
  ierr = PetscOptionsHasName(NULL,NULL,"-x_view",&flg_xview);
  ierr = PetscOptionsGetString(NULL,NULL,"-x_file",xfile,PETSC_MAX_PATH_LEN,&flg_xfile);

  /* Load matrix from file */
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,Ain,FILE_MODE_READ,&fd);CHKERRQ(ierr);
  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);
  ierr = MatSetFromOptions(A);CHKERRQ(ierr);
  ierr = MatLoad(A,fd);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&fd);

  /* Load RHS vector from file */
  ierr = PetscViewerBinaryOpen(PETSC_COMM_WORLD,rhs,FILE_MODE_READ,&fd);CHKERRQ(ierr);
  ierr = VecCreate(PETSC_COMM_WORLD,&b);CHKERRQ(ierr);
  ierr = VecSetFromOptions(b);CHKERRQ(ierr);
  ierr = VecLoad(b,fd);CHKERRQ(ierr);
  ierr = PetscViewerDestroy(&fd);CHKERRQ(ierr);

  /* Linear solver */
  ierr = VecDuplicate(b,&x);CHKERRQ(ierr);
  ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);
  ierr = KSPSetOperators(ksp,A,A);CHKERRQ(ierr);
  ierr = KSPSetFromOptions(ksp);CHKERRQ(ierr);
  ierr = KSPSolve(ksp,b,x);CHKERRQ(ierr);

  /* Postprocessing */
  if (flg_xview) {ierr = VecView(x,PETSC_VIEWER_STDOUT_WORLD);}
  if (flg_xfile) {
    ierr = PetscViewerASCIIOpen(PETSC_COMM_WORLD,xfile,&fd);CHKERRQ(ierr);
    ierr = VecView(x,fd);
    ierr = PetscViewerDestroy(&fd);CHKERRQ(ierr);
  }
  
  /* Cleanup */
  ierr = MatDestroy(&A);CHKERRQ(ierr);
  ierr = VecDestroy(&b);CHKERRQ(ierr);
  ierr = VecDestroy(&x);CHKERRQ(ierr);
  ierr = KSPDestroy(&ksp);CHKERRQ(ierr);
  PetscFinalize();
  return 0;
}
